#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2018 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2018 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import uuid
import heapq

from . import amqp
from . import legacy
from . import exceptions

class Queue(object):

    def length(self):
        raise exceptions.NotImplementedError()

    def clear(self):
        raise exceptions.NotImplementedError()

    def push(self, value, priority = None, identify = False):
        raise exceptions.NotImplementedError()

    def pop(self, block = True, full = False):
        raise exceptions.NotImplementedError()

    def subscribe(self, callback, full = False):
        raise exceptions.NotImplementedError()

    def loop(self):
        raise exceptions.NotImplementedError()

    def unloop(self):
        raise exceptions.NotImplementedError()

    def ack(self):
        raise exceptions.NotImplementedError()

    def nack(self):
        raise exceptions.NotImplementedError()

    def build_value(
        self,
        value,
        priority = None,
        identify = False,
        reverse = False
    ):
        if identify: identifier = self.build_identifier()
        else: identifier = None
        if priority and reverse: priority *= -1
        return (priority, identifier, value), identifier

    def build_identifier(self):
        return str(uuid.uuid4())

class MemoryQueue(Queue):

    def __init__(self):
        Queue.__init__(self)
        self._queue = []

    def length(self):
        return len(self._queue)

    def clear(self):
        del self._queue[:]

    def push(self, value, priority = None, identify = False):
        value, identifier = self.build_value(
            value,
            priority = priority,
            identify = identify,
            reverse = True
        )
        heapq.heappush(self._queue, value)
        return identifier

    def pop(self, block = True, full = False):
        priority, identifier, value = heapq.heappop(self._queue)
        return (priority, identifier, value) if full else value

class MultiprocessQueue(Queue):

    def __init__(self):
        try: import queue
        except ImportError: import Queue as queue
        Queue.__init__(self)
        self._queue = queue.PriorityQueue()

    def length(self):
        return self._queue.qsize()

    def clear(self):
        try: import queue
        except ImportError: import Queue as queue
        self._queue = queue.PriorityQueue()

    def push(self, value, priority = None, identify = False):
        value, identifier = self.build_value(
            value,
            priority = priority,
            identify = identify,
            reverse = True
        )
        self._queue.put(value)
        return identifier

    def pop(self, block = True, full = False):
        priority, identifier, value = self._queue.get(block)
        return (priority, identifier, value) if full else value

class AMQPQueue(Queue):

    def __init__(
        self,
        url = None,
        name = "default",
        durable = False,
        max_priority = 255,
        encoder = "pickle",
        protocol = 2,
        encoding = "utf-8",
        amqp = None
    ):
        self.url = url
        self.name = name
        self.durable = durable
        self.max_priority = max_priority
        self.encoder = encoder
        self.protocol = protocol
        self.encoding = encoding
        self.amqp = amqp
        self._build()

    def clear(self):
        self.channel.queue_purge(queue = self.name)

    def push(self, value, priority = None, identify = False):
        value, identifier = self.build_value(
            value,
            priority = priority,
            identify = identify,
            reverse = False
        )
        body = self._dump(value)
        self.channel.basic_publish(
            exchange = "",
            routing_key = self.name,
            body = body,
            properties = amqp.properties(
                delivery_mode = 2,
                priority = value[0] or 0
            )
        )
        return identifier

    def pop(self, block = True, full = False):
        _method, _properties, body = self.channel.basic_get(
            queue = self.name, no_ack = True
        )
        priority, identifier, value = self._load(body)
        return (priority, identifier, value) if full else value

    def subscribe(self, callback, full = False, no_ack = True):
        def handler(channel, method, properties, body):
            priority, identifier, value = self._load(body)
            result = (priority, identifier, value) if full else value
            ack = lambda: self.ack(delivery_tag = method.delivery_tag)
            nack = lambda: self.ack(delivery_tag = method.delivery_tag)
            callback(result) if no_ack else callback(result, ack, nack)

        self.channel.basic_consume(handler, queue = self.name, no_ack = no_ack)

    def ack(self, delivery_tag = None):
        self.channel.basic_ack(delivery_tag = delivery_tag)

    def nack(self, delivery_tag = None):
        self.channel.basic_nack(delivery_tag = delivery_tag)

    def loop(self):
        self.channel.start_consuming()

    def unloop(self):
        self.channel.stop_consuming()

    def _dump(self, value):
        return self._dumper(value)

    def _dump_pickle(self, value):
        return legacy.cPickle.dumps(value, protocol = self.protocol)

    def _dump_json(self, value):
        body = json.dumps(value)
        return legacy.bytes(body, encoding = self.encoding)

    def _load(self, body):
        return self._loader(body)

    def _load_pickle(self, body):
        if legacy.PYTHON_3: kwargs = dict(encoding = "bytes")
        else: kwargs = dict()
        return legacy.cPickle.loads(body, **kwargs)

    def _load_json(self, body):
        if legacy.is_bytes(body): body = body.decode(self.encoding)
        return json.loads(body)

    def _build(self):
        if not self.amqp: self.amqp = amqp.AMQP(url = self.url)
        self.connection = self.amqp.get_connection()
        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count = 1)
        self.queue = self.channel.queue_declare(
            queue = self.name,
            durable = self.durable,
            arguments = {
                "x-max-priority" : self.max_priority
            }
        )
        self._dumper = getattr(self, "_dump_" + self.encoder)
        self._loader = getattr(self, "_load_" + self.encoder)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2018 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2018 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import shelve
import pickle

from . import common
from . import config
from . import redisdb
from . import component
from . import exceptions

class Preferences(component.Component):

    def __init__(self, name = "preferences", owner = None, *args, **kwargs):
        component.Component.__init__(self, name = name, owner = owner, *args, **kwargs)
        load = kwargs.pop("load", True)
        if load: self.load(*args, **kwargs)

    def __getitem__(self, key):
        return self.get(key, strict = True)

    def __setitem__(self, key, value):
        self.set(key, value)

    def __delitem__(self, key):
        self.delete(key)

    @classmethod
    def new(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def get(self, name, default = None, strict = False, *args, **kwargs):
        return self._get(name, default = default, strict = strict, *args, **kwargs)

    def set(self, name, value, *args, **kwargs):
        flush = kwargs.get("flush", True)
        result = self._set(name, value, *args, **kwargs)
        if flush: self.flush()
        return result

    def delete(self, name, *args, **kwargs):
        return self._delete(name, *args, **kwargs)

    def flush(self, *args, **kwargs):
        return self._flush(*args, **kwargs)

    def clear(self, *args, **kwargs):
        return self._clear(*args, **kwargs)

    def _get(self, name, default = None, strict = False, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def _set(self, name, value, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def _delete(self, name, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def _flush(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def _clear(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

class MemoryPreferences(Preferences):

    def __init__(self, name = "memory", owner = None, *args, **kwargs):
        Preferences.__init__(self, name = name, owner = owner, *args, **kwargs)

    def _load(self, *args, **kwargs):
        Preferences._load(self, *args, **kwargs)
        self._preferences = dict()

    def _unload(self, *args, **kwargs):
        Preferences._unload(self, *args, **kwargs)
        self._preferences = None

    def _get(self, name, default = None, strict = False, *args, **kwargs):
        if strict: return self._preferences[name]
        return self._preferences.get(name, default)

    def _set(self, name, value, *args, **kwargs):
        self._preferences[name] = value

    def _delete(self, name, *args, **kwargs):
        del self._preferences[name]

    def _flush(self, *args, **kwargs):
        pass

    def _clear(self, *args, **kwargs):
        self._preferences.clear()

class FilePreferences(Preferences):

    def __init__(self, name = "file", owner = None, *args, **kwargs):
        Preferences.__init__(self, name = name, owner = owner, *args, **kwargs)

    def _load(self, *args, **kwargs):
        Preferences._load(self, *args, **kwargs)
        self.preferences_path = kwargs.pop("preferences_path", None)
        self._open()

    def _unload(self, *args, **kwargs):
        Preferences._unload(self, *args, **kwargs)
        self._close()

    def _get(self, name, default = None, strict = False, *args, **kwargs):
        if strict: return self._shelve[name]
        return self._shelve.get(name, default)

    def _set(self, name, value, *args, **kwargs):
        self._shelve[name] = value

    def _delete(self, name, *args, **kwargs):
        del self._shelve[name]

    def _flush(self, *args, **kwargs):
        self._sync()

    def _clear(self, *args, **kwargs):
        if not os.path.exists(self.preferences_path): return
        os.remove(self.preferences_path)

    def _open(self):
        self._ensure_path()
        self._shelve = shelve.open(
            self.preferences_path,
            protocol = 2,
            writeback = True
        )

    def _close(self):
        if not self._shelve: return
        self._shelve.close()
        self._shelve = None

    def _ensure_path(self):
        if self.preferences_path: return
        app_path = common.base().get_base_path()
        preferences_path = os.path.join(app_path, "preferences.shelve")
        preferences_path = config.conf("PREFERENCES_PATH", preferences_path)
        preferences_path = os.path.expanduser(preferences_path)
        preferences_path = os.path.abspath(preferences_path)
        preferences_path = os.path.normpath(preferences_path)
        self.preferences_path = preferences_path

    def _sync(self, secure = None):
        if secure == None:
            secure = self._db_secure()
        if secure:
            self._shelve.close()
            self._open()
        else:
            self._shelve.sync()

    def _db_secure(self):
        return self._db_type() == "dbm"

    def _db_type(self):
        shelve_cls = type(self._shelve.dict)
        shelve_dbm = shelve_cls.__name__
        return shelve_dbm

class RedisPreferences(Preferences):

    SERIALIZER = pickle
    """ The serializer to be used for the values contained in
    the preferences (used on top of the class) """

    PREFIX = "appier_preferences_"
    """The default prefix value that is going to
    be prepended to the key name to handle preferences"""

    def __init__(self, name = "redis", owner = None, *args, **kwargs):
        Preferences.__init__(self, name = name, owner = owner, *args, **kwargs)

    def _load(self, *args, **kwargs):
        Preferences._load(self, *args, **kwargs)
        self._serializer = kwargs.pop("serializer", self.__class__.SERIALIZER)
        self._prefix = kwargs.pop("prefix", self.__class__.PREFIX)
        self._prefix += self.owner.name + "_" if self.owner else ""
        self._open()

    def _unload(self, *args, **kwargs):
        Preferences._unload(self, *args, **kwargs)
        self._close()

    def _get(self, name, default = None, strict = False, *args, **kwargs):
        name = self._fqn(name)
        value = self._redis.get(name)
        if not value:
            if strict: raise KeyError("not found")
            return default
        try: return self._serializer.loads(value)
        except: return default

    def _set(self, name, value, *args, **kwargs):
        name = self._fqn(name)
        value = self._serializer.dumps(value)
        self._redis.set(name, value)

    def _delete(self, name, *args, **kwargs):
        name = self._fqn(name)
        self._redis.delete(name)

    def _flush(self, *args, **kwargs):
        pass

    def _clear(self, *args, **kwargs):
        self._redis.flushdb()

    def _open(self):
        self._redis = redisdb.get_connection()
        self._redis.ping()

    def _close(self):
        self._redis = None

    def _fqn(self, name):
        return self._prefix + name



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2018 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2018 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import hmac
import uuid
import time
import base64
import hashlib
import logging

from . import base
from . import http
from . import legacy
from . import observer
from . import exceptions
from . import structures

class API(observer.Observable):
    """
    Abstract and top level API class that should be used
    as the foundation for the creation of API clients.

    This class should offer a set of services so that a
    concrete API implementation should not be concerned
    with issues like: logging, building and destruction.
    """

    SINGLETON = None

    def __init__(self, owner = None, *args, **kwargs):
        observer.Observable.__init__(self, *args, **kwargs)
        self.owner = owner or base.APP
        if not hasattr(self, "auth_callback"): self.auth_callback = None

    @classmethod
    def singleton(cls, *args, **kwargs):
        callback = kwargs.pop("scallback", None)
        if cls.SINGLETON: return cls.SINGLETON
        cls.SINGLETON = cls(*args, **kwargs)
        if callback: callback(cls.SINGLETON)
        return cls.SINGLETON

    def get(
        self,
        url,
        headers = None,
        params = None,
        handle = None,
        silent = None,
        redirect = None,
        timeout = None,
        callback = True,
        extra = None,
        **kwargs
    ):
        headers = headers or dict()
        extra = extra or dict()
        params = structures.OrderedDict(params or kwargs)
        auth_callback = self.auth_callback if callback else None
        self.build(
            "GET",
            url,
            headers = headers,
            params = params,
            kwargs = kwargs
        )
        return self.request(
            http.get,
            url,
            params = params,
            headers = headers,
            handle = handle,
            silent = silent,
            redirect = redirect,
            timeout = timeout,
            auth_callback = auth_callback,
            **extra
        )

    def post(
        self,
        url,
        data = None,
        data_j = None,
        data_m = None,
        headers = None,
        params = None,
        mime = None,
        handle = None,
        silent = None,
        redirect = None,
        timeout = None,
        callback = True,
        extra = None,
        **kwargs
    ):
        headers = headers or dict()
        extra = extra or dict()
        params = structures.OrderedDict(params or kwargs)
        auth_callback = self.auth_callback if callback else None
        self.build(
            "POST",
            url,
            data = data,
            data_j = data_j,
            data_m = data_m,
            headers = headers,
            params = params,
            mime = mime,
            kwargs = kwargs
        )
        return self.request(
            http.post,
            url,
            params = params,
            data = data,
            data_j = data_j,
            data_m = data_m,
            headers = headers,
            mime = mime,
            handle = handle,
            silent = silent,
            redirect = redirect,
            timeout = timeout,
            auth_callback = auth_callback,
            **extra
        )

    def put(
        self,
        url,
        data = None,
        data_j = None,
        data_m = None,
        headers = None,
        params = None,
        mime = None,
        handle = None,
        silent = None,
        redirect = None,
        timeout = None,
        callback = True,
        extra = None,
        **kwargs
    ):
        headers = headers or dict()
        extra = extra or dict()
        params = structures.OrderedDict(params or kwargs)
        auth_callback = self.auth_callback if callback else None
        self.build(
            "PUT",
            url,
            data = data,
            data_j = data_j,
            data_m = data_m,
            headers = headers,
            params = params,
            mime = mime,
            kwargs = kwargs
        )
        return self.request(
            http.put,
            url,
            params = params,
            data = data,
            data_j = data_j,
            data_m = data_m,
            headers = headers,
            mime = mime,
            handle = handle,
            silent = silent,
            redirect = redirect,
            timeout = timeout,
            auth_callback = auth_callback,
            **extra
        )

    def delete(
        self,
        url,
        headers = None,
        params = None,
        handle = None,
        silent = None,
        redirect = None,
        timeout = None,
        callback = True,
        extra = None,
        **kwargs
    ):
        headers = headers or dict()
        extra = extra or dict()
        params = structures.OrderedDict(params or kwargs)
        auth_callback = self.auth_callback if callback else None
        self.build(
            "DELETE",
            url,
            headers = headers,
            params = params,
            kwargs = kwargs
        )
        return self.request(
            http.delete,
            url,
            params = params,
            headers = headers,
            handle = handle,
            silent = silent,
            redirect = redirect,
            timeout = timeout,
            auth_callback = auth_callback,
            **extra
        )

    def patch(
        self,
        url,
        data = None,
        data_j = None,
        data_m = None,
        headers = None,
        params = None,
        mime = None,
        handle = None,
        silent = None,
        redirect = None,
        timeout = None,
        callback = True,
        extra = None,
        **kwargs
    ):
        headers = headers or dict()
        extra = extra or dict()
        params = structures.OrderedDict(params or kwargs)
        auth_callback = self.auth_callback if callback else None
        self.build(
            "PATCH",
            url,
            data = data,
            data_j = data_j,
            data_m = data_m,
            headers = headers,
            params = params,
            mime = mime,
            kwargs = kwargs
        )
        return self.request(
            http.patch,
            url,
            params = params,
            data = data,
            data_j = data_j,
            data_m = data_m,
            headers = headers,
            mime = mime,
            handle = handle,
            silent = silent,
            redirect = redirect,
            timeout = timeout,
            auth_callback = auth_callback,
            **extra
        )

    def request(self, method, *args, **kwargs):
        try: result = method(*args, **kwargs)
        except exceptions.HTTPError as exception:
            self.handle_error(exception)
        else: return result

    def build(
        self,
        method,
        url,
        data = None,
        data_j = None,
        data_m = None,
        headers = None,
        params = None,
        mime = None,
        kwargs = None
    ):
        pass

    def handle_error(self, error):
        raise

    @property
    def logger(self):
        if self.owner: return self.owner.logger
        else: return logging.getLogger()

class OAuthAPI(API):
    """
    Abstract class for the implementation of oauth based
    infra-structure, should expose a series of utilities
    for the authentication using oauth.

    A direct fallback mode should be support so that dual
    API clients are allowed and may be able to interact.
    """

    DIRECT_MODE = 1
    """ The direct mode where a complete access is allowed
    to the client by providing the "normal" credentials to
    it and ensuring a complete authentication """

    OAUTH_MODE = 2
    """ The oauth client mode where the set of permissions
    (scope) is authorized on behalf on an already authenticated
    user using a web agent (recommended mode) """

    UNSET_MODE = 3
    """ The unset client mode for situations where the client
    exists but not enough information is provided to it so that
    it knows how to interact with the server side (detached client) """

    def __init__(self, *args, **kwargs):
        API.__init__(self, *args, **kwargs)
        self.mode = OAuthAPI.OAUTH_MODE

    def handle_error(self, error):
        raise exceptions.OAuthAccessError(
            message = "Problems using access token found must re-authorize",
            original = error
        )

    def is_direct(self):
        return self.mode == OAuthAPI.DIRECT_MODE

    def is_oauth(self):
        return self.mode == OAuthAPI.OAUTH_MODE

    def _get_mode(self):
        return OAuthAPI.OAUTH_MODE

class OAuth1API(OAuthAPI):

    def __init__(self, *args, **kwargs):
        OAuthAPI.__init__(self, *args, **kwargs)
        self.oauth_token = None
        self.oauth_token_secret = None

    def build(
        self,
        method,
        url,
        data = None,
        data_j = None,
        data_m = None,
        headers = None,
        params = None,
        mime = None,
        kwargs = None
    ):
        if not self.is_oauth(): return
        auth = kwargs.pop("auth", True)
        if auth: self.auth_header(method, url, headers, kwargs)

    def auth_header(self, method, url, headers, kwargs, sign_method = "HMAC-SHA1"):
        if not sign_method == "HMAC-SHA1": raise exceptions.NotImplementedError()

        if not self.client_key: raise exceptions.SecurityError(
            message = "No client key defined, mandatory"
        )
        if not self.client_secret: raise exceptions.SecurityError(
            message = "No client secret defined, mandatory"
        )

        oauth_callback = kwargs.pop("oauth_callback", None)

        params = []

        encoded = http._quote(kwargs, safe = "~")
        items = encoded.items()
        params.extend(items)

        unique = str(uuid.uuid4())
        oauth_nonce = unique.replace("-", "")
        oauth_timestamp = str(int(time.time()))

        authorization = dict(
            oauth_nonce = oauth_nonce,
            oauth_signature_method = sign_method,
            oauth_timestamp = oauth_timestamp,
            oauth_consumer_key = self.client_key,
            oauth_version = "1.0"
        )
        if self.oauth_token: authorization["oauth_token"] = self.oauth_token
        if oauth_callback: authorization["oauth_callback"] = oauth_callback

        encoded = http._quote(authorization, safe = "~")
        items = encoded.items()
        params.extend(items)
        params.sort()

        signature_base = "&".join(["%s=%s" % (key, value) for key, value in params])
        signature_extra = "&".join([
            legacy.quote(method, safe = "~"),
            legacy.quote(url, safe = "~"),
            legacy.quote(signature_base, safe = "~")
        ])

        if self.oauth_token_secret: key = "%s&%s" % (self.client_secret, self.oauth_token_secret)
        else: key = "%s&" % self.client_secret

        if legacy.is_unicode(key): key = key.encode("utf-8")
        if legacy.is_unicode(signature_extra): signature_extra = signature_extra.encode("utf-8")

        oauth_signature = hmac.new(key, signature_extra, hashlib.sha1).digest()
        oauth_signature = base64.b64encode(oauth_signature)
        oauth_signature = legacy.str(oauth_signature)

        authorization["oauth_signature"] = oauth_signature
        authorization = http._quote(authorization, safe = "~")
        authorization = authorization.items()
        authorization = ["%s=\"%s\"" % (key, value) for key, value in authorization]

        authorization_s = ", ".join(authorization)
        authorization_s = "OAuth %s" % authorization_s

        headers["Authorization"] = authorization_s

class OAuth2API(OAuthAPI):

    def __init__(self, *args, **kwargs):
        OAuthAPI.__init__(self, *args, **kwargs)
        self.access_token = None

    def build(
        self,
        method,
        url,
        data = None,
        data_j = None,
        data_m = None,
        headers = None,
        params = None,
        mime = None,
        kwargs = None
    ):
        if not self.is_oauth(): return
        token = kwargs.pop("token", True)
        if token and "param" in self.oauth_types:
            kwargs[self.oauth_param] = self.get_access_token()
        if token and "header" in self.oauth_types:
            headers["Authorization"] = "Bearer %s" % self.get_access_token()

    def get_access_token(self):
        if self.access_token: return self.access_token
        raise exceptions.OAuthAccessError(
            message = "No access token found must re-authorize"
        )

    @property
    def oauth_types(self):
        return ("param", "header")

    @property
    def oauth_param(self):
        return "access_token"
